import json
from typing import Any, Callable, TypeVar, Union, Dict

from .constants import CONTENT_TYPE

_V = TypeVar("_V")


MediaHandler = Callable[[Any], str]
Handlers = Dict[str, MediaHandler]


def handle_json(value: Union[dict, list]) -> str:
    """A media handler that dumps a value using `json.dumps`."""
    return json.dumps(value)


def get_default_handlers() -> Handlers:
    """Return the default media handlers.

    - `application/json`: [handle_json](#handle-json)
    """
    return {CONTENT_TYPE.JSON: handle_json}


class UnsupportedMediaType(Exception):
    """Raised when trying to use an unsupported media type.

    # Attributes
    media_type (str): the unsupported media type.
    handlers (dict): the dict of supported media handlers.
    """

    def __init__(self, media_type: str, handlers: Handlers):
        self.media_type = media_type
        self.handlers = handlers
        self._available = ", ".join(handlers.keys())

    def __str__(self):
        return f"{self.media_type} (available: {self._available})"



from .applications import API, App
from .errors import HTTPError
from .middleware import Middleware
from .recipes import Recipe
from .response import Response
from .request import Request
from .staticfiles import static
from .templates import Templates
from .views import view
from .websockets import WebSocket, WebSocketDisconnect

__version__ = "0.12.1"



import inspect
from functools import wraps
from typing import Callable, cast, Dict, Union, Awaitable, Type

from .compat import call_async
from .request import Request
from .response import Response
from .routing import HTTPRoute
from .views import Handler, get_handlers, View

HookFunction = Callable[[Request, Response, dict], Awaitable[None]]
HookCollection = Dict[HTTPRoute, HookFunction]

BEFORE = "before"
AFTER = "after"


class Hooks:
    """Hooks manager."""

    def before(self, hook: HookFunction, *args, **kwargs):
        """Register a before hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(BEFORE, hook, *args, **kwargs)

    def after(self, hook: HookFunction, *args, **kwargs):
        """Register an after hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(AFTER, hook, *args, **kwargs)

    def _hook_decorator(
        self, hook_type: str, hook: HookFunction, *args, **kwargs
    ):
        # Enclose args and kwargs
        async def hook_func(req: Request, res: Response, params: dict):
            await call_async(  # type: ignore
                hook, req, res, params, *args, **kwargs
            )

        def decorator(handler: Union[Type[View], Handler]):
            """Attach the hook to the given handler."""
            if not inspect.isclass(handler):
                return _with_hook(hook_type, hook_func, cast(Handler, handler))

            view_cls = cast(View, handler)

            # Recursively apply hook to all handlers.
            for method, _handler in get_handlers(view_cls).items():
                setattr(view_cls, method, decorator(_handler))

            return view_cls

        return decorator


def _with_hook(hook_type: str, func: HookFunction, handler: Handler):
    async def call_hook(args, kw):
        if len(args) == 2:
            req, res = args
        else:
            # method that has `self` as a first parameter
            req, res = args[1:3]
        assert isinstance(req, Request)
        assert isinstance(res, Response)
        await call_async(func, req, res, kw)

    if hook_type == BEFORE:

        @wraps(handler)
        async def with_before_hook(*args, **kwargs):
            await call_hook(args, kwargs)
            await call_async(handler, *args, **kwargs)

        return with_before_hook

    assert hook_type == AFTER

    @wraps(handler)
    async def with_after_hook(*args, **kwargs):
        await call_async(handler, *args, **kwargs)
        await call_hook(args, kwargs)

    return with_after_hook


# Pre-bind to module
hooks = Hooks()
before = hooks.before
after = hooks.after



from .request import Request
from .response import Response
from .errors import HTTPError


# Built-in HTTP error handlers.


async def error_to_html(req: Request, res: Response, exc: HTTPError):
    """Convert an exception to an HTML response.

    The response contains a `<h1>` tag with the error's `title` and,
    if provided, a `<p>` tag with the error's `detail`.

    # Example

    ```html
    <h1>403 Forbidden</h1>
    <p>You do not have the permissions to perform this operation.</p>
    ```
    """
    res.status_code = exc.status_code
    html = f"<h1>{exc.title}</h1>"
    if exc.detail:
        html += f"\n<p>{exc.detail}</p>"
    res.html = html


async def error_to_media(req: Request, res: Response, exc: HTTPError):
    """Convert an exception to a media response.

    The response contains the following items:

    - `error`: the error's `title`
    - `status`: the error's `status_code`
    - `detail`: the error's `detail` (if provided)

    # Example

    ```json
    {
        "error": "403 Forbidden",
        "status": 403,
        "detail": "You do not have the permissions to perform this operation."
    }
    ```
    """
    res.status_code = exc.status_code
    media = {"error": exc.title, "status": exc.status_code}
    if exc.detail:
        media["detail"] = exc.detail
    res.media = media


async def error_to_text(req: Request, res: Response, exc: HTTPError):
    """Convert an exception to a plain text response.

    The response contains a line with the error's `title` and, if provided,
    a line for the error's `detail`.

    # Example
    ```
    403 Forbidden
    You do not have the permissions to perform this operation.
    ```
    """
    res.status_code = exc.status_code
    text = exc.title
    if exc.detail:
        text += f"\n{exc.detail}"
    res.text = text



import inspect
from os.path import basename
from typing import (
    Any,
    AsyncIterable,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Union,
)

from starlette.background import BackgroundTask
from starlette.requests import Request
from starlette.responses import (
    Response as _Response,
    StreamingResponse as _StreamingResponse,
    FileResponse as _FileResponse,
)

from .constants import CONTENT_TYPE
from .media import MediaHandler

AnyStr = Union[str, bytes]
BackgroundFunc = Callable[..., Coroutine]
Stream = AsyncIterable[AnyStr]
StreamFunc = Callable[[], Stream]


def _content_setter(content_type: str):
    def fset(res: "Response", value: AnyStr):
        res.content = value
        res.headers["content-type"] = content_type

    doc = (
        "Write-only property that sets `content` to the set value and sets "
        f'the `Content-Type` header to `"{content_type}"`.'
    )

    return property(fget=None, fset=fset, doc=doc)


class Response:
    """The response builder, passed to HTTP views and typically named `res`.

    At the lower-level, `Response` behaves like an ASGI application instance:
    it is a callable and accepts `receive` and `send` as defined in the [ASGI
    spec](https://asgi.readthedocs.io/en/latest/specs/main.html#applications).

    [media]: ../guides/http/media.md
    [Content-Disposition]: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition

    # Parameters
    request (Request): the currently processed request.
    media_type (str): the configured media type (given by the `App`).
    media_handler (callable): the configured media handler (given by the `App`).

    # Attributes
    content (bytes or str): the raw response content.
    status_code (int): the HTTP status code. If not set, defaults to `200`.
    headers (dict):
        a case-insensitive dictionary of HTTP headers.
        If not set, `content-type` header is set to `text/plain`.
    chunked (bool): sets the `transfer-encoding` header to `chunked`.
    attachment (str):
        a file name that this response should be sent as.
        This is done by setting the [Content-Disposition] header, and
        typically makes the client browser trigger a "Save As…" dialog or
        download and save the file locally.
    """

    text = _content_setter(CONTENT_TYPE.PLAIN_TEXT)
    html = _content_setter(CONTENT_TYPE.HTML)

    def __init__(
        self, request: Request, media_type: str, media_handler: MediaHandler
    ):
        # Public attributes.
        self.content: Optional[AnyStr] = None
        self.request = request
        self.status_code: Optional[int] = None
        self.headers: Dict[str, str] = {}
        self.chunked = False
        self.attachment: Optional[str] = None
        # Private attributes.
        self._file_path: Optional[str] = None
        self._media_type = media_type
        self._media_handler = media_handler
        self._background: Optional[BackgroundFunc] = None
        self._stream: Optional[Stream] = None

    media = property(
        doc=(
            "Write-only property that sets `content` to the set value "
            "serializer using the `media_handler`, sets the "
            "`Content-Type` header to the `media_type`."
        )
    )

    @media.setter
    def media(self, value: Any):
        self.content = self._media_handler(value)
        self.headers["content-type"] = self._media_type

    def file(self, path: str, attach: bool = True):
        """Send a file asynchronously using [aiofiles].

        [aiofiles]: https://github.com/Tinche/aiofiles

        # Parameters
        path (str):
            A path to a file on this machine.
        attach (bool, optional):
            Whether to send the file as an [attachment](#response).
            Defaults to `True`.
        """
        self._file_path = path
        if attach:
            self.attachment = basename(path)

    def background(
        self, func: BackgroundFunc, *args, **kwargs
    ) -> BackgroundFunc:
        """Register a coroutine function to be executed in the background.

        This can be used either as a decorator or a regular function.

        # Parameters
        func (callable):
            A coroutine function.
        *args, **kwargs:
            Any positional and keyword arguments to pass to `func` when
            executing it.
        """

        async def background():
            await func(*args, **kwargs)

        self._background = background
        return func

    @property
    def _background_task(self) -> Optional[BackgroundTask]:
        if self._background is not None:
            return BackgroundTask(self._background)
        return None

    def stream(self, func: StreamFunc) -> StreamFunc:
        """Stream the response.

        Should be used to decorate a no-argument asynchronous generator
        function.
        """
        assert inspect.isasyncgenfunction(func)
        self._stream = func()
        return func

    async def __call__(self, receive, send):
        """Build and send the response."""
        if self.status_code is None:
            self.status_code = 200

        if self.status_code != 204:
            self.headers.setdefault("content-type", "text/plain")

        if self.chunked:
            self.headers["transfer-encoding"] = "chunked"

        if self.attachment is not None:
            disposition = f"attachment; filename='{self.attachment}'"
            self.headers.setdefault("content-disposition", disposition)

        response_kwargs = {
            "content": self.content,
            "headers": self.headers,
            "status_code": self.status_code,
            "background": self._background_task,
        }

        response_cls = _Response

        if self._file_path is not None:
            response_cls = _FileResponse
            response_kwargs["path"] = self._file_path
            # `FileResponse` will populate the response from `path` and
            # doesn't expect `content` to be passed.
            del response_kwargs["content"]
            # `FileResponse` will set the status code to 200.
            del response_kwargs["status_code"]
        elif self._stream is not None:
            response_cls = _StreamingResponse
            response_kwargs["content"] = self._stream

        response: _Response = response_cls(**response_kwargs)
        await response(receive, send)



"""This module defines classes (both abstract and concrete) that power routing.

::: tip NOTATIONS
This modules uses the following [generic types]:

- `_R` refers to a route object, i.e. an instance of a subclass of [BaseRoute](#baseroute).
- `_V` refers to a view object.

[generic types]: https://docs.python.org/3/library/typing.html#generics
:::
"""

import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    NoReturn,
    Optional,
    Tuple,
    TypeVar,
)

from parse import Parser
from starlette.websockets import WebSocketClose

from . import views
from .app_types import HTTPApp, Receive, Scope, Send
from .errors import HTTPError
from .redirection import Redirection
from .request import Request
from .response import Response
from .views import AsyncHandler, HandlerDoesNotExist, View
from .websockets import WebSocket, WebSocketView

WILDCARD = "{}"

# Route generic types.
_R = TypeVar("_R", bound="BaseRoute")  # route
_V = TypeVar("_V")  # view

# Base classes.


class BaseRoute(Generic[_V]):
    """The base route class.

    This is referenced as `_R` in the rest of this module.

    # Parameters
    pattern (str): an URL pattern.
    view (_V):
        a view function or object whose actual type is defined by concrete
        routes.
    """

    def __init__(self, pattern: str, view: _V):
        if pattern != WILDCARD and not pattern.startswith("/"):
            pattern = f"/{pattern}"
        self._pattern = pattern
        self._parser = Parser(self._pattern)
        self.view = view

    @property
    def pattern(self) -> str:
        return self._pattern

    def url(self, **kwargs) -> str:
        """Return the full URL path for the given route parameters.

        # Parameters
        kwargs (dict): route parameters.

        # Returns
        url (str):
            A full URL path obtained by formatting the route pattern with
            the provided route parameters.
        """
        return self._pattern.format(**kwargs)

    def parse(self, path: str) -> Optional[dict]:
        """Parse an URL path against the route's URL pattern.

        # Returns
        params (dict or None):
            If the URL path matches the URL pattern, this is a dictionary
            containing the route parameters, otherwise it is `None`.
        """
        result = self._parser.parse(path)
        return result.named if result is not None else None


class RouteMatch(Generic[_R]):  # pylint: disable=unsubscriptable-object
    """Represents a match between an URL path and a route.

    # Parameters
    route (BaseRoute): a route object.
    params (dict): extracted route parameters.
    """

    def __init__(self, route: _R, params: dict):
        self.route = route
        self.params = params


class BaseRouter(Generic[_R, _V]):
    """The base router class.

    # Attributes
    routes (dict):
        A mapping of URL patterns to route objects.
    """

    def __init__(self):
        self.routes: Dict[str, _R] = {}

    def _get_key(self, route: _R) -> str:
        # Return the key at which `route` should be stored internally.
        raise NotImplementedError

    def normalize(self, view: Any) -> _V:
        """Perform any conversion necessary to return a proper view object.

        This is a no-op by default, i.e. it returns what it's given.
        """
        return view

    def add_route(self, view: _V, pattern: str, **kwargs) -> _R:
        """Register a route (to be implemented by concrete routers)."""
        raise NotImplementedError

    def add(self, route: _R) -> None:
        self.routes[self._get_key(route)] = route

    def route(self, *args, **kwargs) -> Callable[[Any], _R]:
        """Register a route by decorating a view.

        The decorated function or class will be converted to a proper view using
        [`.normalize()`](#normalize), and then fed to
        [`.add_route()`](#add-route).

        # Parameters
        *args, **kwargs:
            passed to [`.add_route()`](#add-route)
            along with the normalized view.
        """

        def decorate(view: Any) -> _R:
            normalized_view: _V = self.normalize(view)
            return self.add_route(normalized_view, *args, **kwargs)

        return decorate

    def match(self, path: str) -> Optional[RouteMatch[_R]]:
        """Attempt to match an URL path against one of the registered routes.

        # Parameters
        path (str): an URL path

        # Returns
        match (RouteMatch or None):
            a [`RouteMatch`](#routematch) object if the path matched
            a registered route, `None` otherwise.
        """
        for route in self.routes.values():
            params = route.parse(path)
            if params is not None:
                return RouteMatch(route=route, params=params)
        return None


# HTTP.


class HTTPRoute(BaseRoute[View]):
    """Represents the binding of an URL pattern to an HTTP view.

    Subclass of [BaseRoute](#baseroute).

    # Parameters
    pattern (str): an URL pattern.
    view (View):
        A `View` object.
    name (str):
        The route's name.
    """

    def __init__(self, pattern: str, view: View, name: str):
        super().__init__(pattern, view)
        self.name = name

    async def __call__(self, req: Request, res: Response, **params):
        method: str = req.method.lower()

        try:
            handler: AsyncHandler = self.view.get_handler(method)
        except HandlerDoesNotExist as e:
            raise HTTPError(405) from e

        await handler(req, res, **params)  # type: ignore


class HTTPRouter(HTTPApp, BaseRouter[HTTPRoute, View]):
    """A router for HTTP routes.

    Subclass of [BaseRouter](#baserouter).

    Note: routes are stored by `name` instead of `pattern`.
    """

    def _get_key(self, route: HTTPRoute) -> str:
        # NOTE: this ensures that no two routes stored in this
        # router have the same name.
        return route.name

    def normalize(self, view: Any) -> View:
        """Build a `View` object.

        The input, free-form `view` object is converted using the following
        rules:

        - Classes are instanciated (without arguments) and converted with
        [`from_obj()`][obj].
        - Callables are converted with [`from_handler()`][handler].
        - Any other object is interpreted as a view-like object, and converted
        with [`from_obj()`][obj].

        [handler]: ./views.md#from-handler
        [obj]: ./views.md#from-obj

        # Returns
        view (View):
            a `View` object, ready to be fed to [`.add_route()`](#add-route).
        """

        if isinstance(view, View):
            return view

        if inspect.isclass(view):
            # View-like class.
            return views.from_obj(view())

        if callable(view):
            # Function-based view.
            # NOTE: here, we ensure backward-compatibility with the routing of
            # function-based views pre-0.9.
            return views.from_handler(view)

        # Treat as a view-like object.
        return views.from_obj(view)

    def add_route(
        self,
        view: View,
        pattern: str,
        name: str = None,
        namespace: str = None,
        **kwargs,
    ) -> HTTPRoute:
        """Register an HTTP route.

        # Parameters
        view (View):
            a `View` object. You may use [.normalize()](#normalize-2)
            to get one from a function or class-based view before-hand.
        pattern (str): an URL pattern.
        name (str): a route name (inferred from the view if not given).
        namespace (str): an optional route namespace.

        # Returns
        route (HTTPRoute): the registered route.
        """
        if name is None:
            name = view.name
        if namespace is not None:
            name = namespace + ":" + name

        route = HTTPRoute(pattern=pattern, view=view, name=name)
        self.add(route)

        return route

    async def __call__(self, req: Request, res: Response) -> Response:
        match = self.match(req.url.path)

        if match is None:
            raise HTTPError(status=404)

        try:
            await match.route(req, res, **match.params)
        except Redirection as redirection:
            res = redirection.response

        return res


# WebSocket.


class WebSocketRoute(BaseRoute[WebSocketView]):
    """Represents the binding of an URL pattern to a WebSocket view.

    Subclass of [BaseRoute](#baseroute).

    [WebSocket]: ./websockets.md#websocket

    # Parameters
    pattern (str): an URL pattern.
    view (coroutine function):
        Should take as parameters a `WebSocket` object and
        any extracted route parameters.
    kwargs (any): passed when building the [WebSocket] object.
    """

    def __init__(self, pattern: str, view: WebSocketView, **kwargs):
        super().__init__(pattern, view)
        self._ws_kwargs = kwargs

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send, **params
    ):
        ws = WebSocket(scope, receive=receive, send=send, **self._ws_kwargs)
        try:
            await self.view(ws, **params)  # type: ignore
        except BaseException:
            await ws.ensure_closed(1011)
            raise


class WebSocketRouter(BaseRouter[WebSocketRoute, WebSocketView]):
    """A router for WebSocket routes.

    Subclass of [BaseRouter](#baserouter).
    """

    def _get_key(self, route: WebSocketRoute) -> str:
        return route.pattern

    def add_route(
        self, view: WebSocketView, pattern: str, **kwargs
    ) -> WebSocketRoute:
        """Register a WebSocket route.

        # Parameters
        pattern (str): an URL pattern.
        view (coroutine function): a WebSocket view.

        # Returns
        route (WebSocketRoute): the registered route.
        """
        route = WebSocketRoute(pattern=pattern, view=view, **kwargs)
        self.add(route)
        return route

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # Dispatch a WebSocket connection request.
        match = self.match(scope["path"])
        if match is None:
            # Close with a 403 error code, as specified in the ASGI spec:
            # https://asgi.readthedocs.io/en/latest/specs/www.html#close
            await WebSocketClose(code=403)(receive, send)
        else:
            await match.route(scope, receive, send, **match.params)


class RoutingMixin:
    """Provide HTTP and WebSocket routing to an application class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.http_router = HTTPRouter()
        self.websocket_router = WebSocketRouter()

    def route(self, pattern: str, *, name: str = None, namespace: str = None):
        """Register a new route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.
        methods (list of str):
            An optional list of HTTP methods.
            Defaults to `["get", "head"]`.
            Ignored for class-based views.
        name (str):
            An optional name for the route.
            If a route already exists for this name, it is replaced.
            Defaults to a snake-cased version of the view's name.
        namespace (str):
            An optional namespace for the route. If given, it is prefixed to
            the name and separated by a colon.
        """
        return self.http_router.route(
            pattern=pattern, name=name, namespace=namespace
        )

    def websocket_route(
        self,
        pattern: str,
        *,
        value_type: Optional[str] = None,
        receive_type: Optional[str] = None,
        send_type: Optional[str] = None,
        caught_close_codes: Optional[Tuple[int]] = None,
    ):
        """Register a WebSocket route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.

        # See Also
        - [WebSocket](./websockets.md#websocket) for a description of keyword
        arguments.
        """
        # NOTE: use named keyword arguments instead of `**kwargs` to improve
        # their accessibility (e.g. for IDE discovery).
        return self.websocket_router.route(
            pattern=pattern,
            value_type=value_type,
            receive_type=receive_type,
            send_type=send_type,
            caught_close_codes=caught_close_codes,
        )

    def url_for(self, name: str, **kwargs) -> str:
        """Build the full URL path for a named route.

        # Parameters
        name (str): the name of the route.
        kwargs (dict): route parameters.

        # Returns
        url (str): an URL path.

        # Raises
        HTTPError(404) : if no route exists for the given `name`.
        """
        route = self.http_router.routes.get(name)
        if route is None:
            raise HTTPError(404)
        return route.url(**kwargs)

    def redirect(
        self,
        *,
        name: str = None,
        url: str = None,
        permanent: bool = False,
        **kwargs,
    ) -> NoReturn:
        """Redirect to another HTTP route.

        This is only meant to be used inside an HTTP view.

        # Parameters
        name (str): name of the route to redirect to.
        url (str):
            URL of the route to redirect to (required if `name` is omitted).
        permanent (bool):
            If `False` (the default), returns a temporary redirection (302).
            If `True`, returns a permanent redirection (301).
        kwargs (dict):
            Route parameters.

        # Raises
        Redirection:
            an exception that will be caught to trigger a redirection.

        # See Also
        - [Redirecting](../guides/http/redirecting.md)
        """
        if name is not None:
            url = self.url_for(name=name, **kwargs)
        else:
            assert url is not None, "url is expected if no route name is given"
        raise Redirection(url=url, permanent=permanent)



from typing import Awaitable, Callable, Optional, Any, Union, Tuple

from starlette.datastructures import URL
from starlette.websockets import (
    WebSocket as StarletteWebSocket,
    WebSocketDisconnect as _WebSocketDisconnect,
)

from .app_types import Event, Scope, Receive, Send
from .constants import WEBSOCKET_CLOSE_CODES


class WebSocket:
    """Represents a WebSocket connection.

    # Parameters
    value_type (str):
        The type of messages received or sent over the WebSocket.
        If given, overrides `receive_type` and `send_type`.
        Defaults to `None`.
    receive_type (str):
        The type of messages received over the WebSocket.
        Defaults to `"text"`.
    send_type (str):
        The type of messages send over the WebSocket.
        Defaults to `"text"`.
    caught_close_codes (tuple of int):
        Close codes of `WebSocketDisconnect` exceptions that should be
        caught and silenced. Defaults to `(1000, 1001)`.
    args (any):
        Passed to the underlying Starlette `WebSocket` object. This is
        typically the ASGI `scope`, `receive` and `send` objects.
    """

    __default_receive_type__ = "text"
    __default_send_type__ = "text"

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        value_type: Optional[str] = None,
        receive_type: Optional[str] = None,
        send_type: Optional[str] = None,
        caught_close_codes: Optional[Tuple[int, ...]] = None,
    ):
        # NOTE: we use composition over inheritance here, because
        # we want to redefine `receive()` and `send()` but Starlette's
        # WebSocket class uses those in many other functions, which we
        # do not need / want to re-implement.
        # The compromise is the definition of delegated methods below.
        self._ws = StarletteWebSocket(scope, receive=receive, send=send)

        if caught_close_codes is None:
            caught_close_codes = (1000, 1001)
        if caught_close_codes is all:
            caught_close_codes = tuple(WEBSOCKET_CLOSE_CODES)
        self.caught_close_codes = caught_close_codes

        if value_type is not None:
            receive_type = send_type = value_type
        else:
            receive_type = receive_type or self.__default_receive_type__
            send_type = send_type or self.__default_send_type__

        self.receive_type = receive_type
        self.send_type = send_type

    @property
    def url(self) -> URL:
        """The URL which the WebSocket connection was made to.

        # See Also
        - [starlette.WebSocket.url](https://www.starlette.io/websockets/#url)
        """
        return self._ws.url

    # Connection handling.

    async def accept(self, subprotocol: str = None) -> None:
        """Accept the connection request.

        # Parameters
        subprotocol (str): a specific WebSocket subprotocol that should be expected from clients. Defaults to `None`.
        """
        return await self._ws.accept(subprotocol=subprotocol)

    async def reject(self):
        """Reject the connection request.

        This is equivalent to `.close(403)`.
        Calling this before `.accept()` has undefined behavior.
        """
        return await self.close(code=403)

    async def close(self, code: int = 1000) -> None:
        """Close the WebSocket connection.

        # Parameters
        code (int): a close code, defaults to `1000` (Normal Closure).

        # See Also
        - [CloseEvent specification](https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent)
        """
        return await self._ws.close(code=code)

    async def ensure_closed(self, code: int = 1000):
        """Close the connection if it has not been closed already."""
        try:
            await self.close(code=code)
        except RuntimeError:
            # Already closed.
            pass

    # Data reception/sending.

    async def receive_text(self) -> str:
        """Receive a message as text."""
        return await self._ws.receive_text()

    async def send_text(self, data: str):
        """Send a message as text."""
        return await self._ws.send_text(data)

    async def receive_bytes(self) -> bytes:
        """Receive a message a bytes."""
        return await self._ws.receive_bytes()

    async def send_bytes(self, data: bytes):
        """Send a message as bytes."""
        return await self._ws.send_bytes(data)

    async def receive_json(self) -> Union[dict, list]:
        """Receive a message as text and parse it to a JSON object.

        # Raises
        json.JSONDecodeError: if the received JSON is invalid.
        """
        return await self._ws.receive_json()

    async def send_json(self, data: Union[dict, list]):
        """Serialize an object to JSON and send it as text.

        # Raises
        TypeError: if the given `data` is not JSON serializable.
        """
        return await self._ws.send_json(data)

    async def receive_event(self) -> Event:
        """Receive a raw ASGI event."""
        return await self._ws.receive()

    async def send_event(self, event: Event):
        """Send a raw ASGI event."""
        return await self._ws.send(event)

    async def receive(self) -> Union[str, bytes, list, dict]:
        """Receive a message from the WebSocket.

        Shortcut for `receive_<self.receive_type>`.
        """
        receiver = getattr(self, f"receive_{self.receive_type}")
        return await receiver()

    async def send(self, message: Any):
        """Send a message over the WebSocket.

        Shortcut for `send_<self.send_type>`.
        """
        sender = getattr(self, f"send_{self.send_type}")
        return await sender(message)

    # Asynchronous context manager.

    async def __aenter__(self, *args, **kwargs):
        await self.accept()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type == WebSocketDisconnect:
            # Client has closed the connection.
            # Returning `True` here silences the exception. See:
            # https://docs.python.org/3/reference/datamodel.html#object.__exit__
            return exc_val.code in self.caught_close_codes

        # Close with Internal Error if an unknown exception was raised.
        code = 1000 if exc_type is None else 1011
        # NOTE: view may have closed the WebSocket already, so we
        # must use ensure_closed().
        await self.ensure_closed(code)

    # Asynchronous iterator.

    async def __aiter__(self):
        while True:
            yield await self.receive()


WebSocketView = Callable[[WebSocket], Awaitable[None]]
WebSocketDisconnect = _WebSocketDisconnect



from typing import Dict, Tuple

ALL_HTTP_METHODS: Tuple[str, ...] = (
    "GET",
    "HEAD",
    "POST",
    "PUT",
    "DELETE",
    "OPTIONS",
    "PATCH",
)

DEFAULT_CORS_CONFIG = {"allow_origins": [], "allow_methods": ["GET"]}

# See: https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent
WEBSOCKET_CLOSE_CODES: Dict[int, str] = {
    1000: "Normal Closure",
    1001: "Going Away",
    1002: "Protocol Error",
    1003: "Unsupported Type",
    # 1004 is reserved
    1005: "No Status Code [Internal]",
    1006: "Connection Closed Abnormally [Internal]",
    1007: "Invalid Data",
    1008: "Policy Violation",
    1009: "Message Too Big",
    1010: "Extension Required",
    1011: "Internal Error",
    1015: "TLS Failure [Internal]",
}


class CONTENT_TYPE:
    PLAIN_TEXT = "text/plain"
    HTML = "text/html"
    JSON = "application/json"



from typing import Optional

from .app_types import HTTPApp
from .compat import call_async
from .request import Request
from .response import Response


class Middleware(HTTPApp):
    """Base class for middleware classes.

    # Parameters
    app: a callable that may as well be another `Middleware` instance.
    kwargs (any):
        Keyword arguments passed when registering the
        middleware on the application.
    """

    def __init__(self, app: HTTPApp, **kwargs):
        self.app = app
        self.kwargs = kwargs

    async def before_dispatch(
        self, req: Request, res: Response
    ) -> Optional[Response]:
        """Perform processing before a request is dispatched.

        If the `Response` object is returned, it will be used
        and no further processing will be performed.

        # Parameters
        req (Request): a Request object.
        res (Response): a Response object.

        # Returns
        res (Response or None): an optional response object.
        """

    async def after_dispatch(
        self, req: Request, res: Response
    ) -> Optional[Response]:
        """Perform processing after a request has been dispatched.

        # Parameters
        req (Request): a Request object.
        res (Response): a Response object.

        # Returns
        res (Response or None): an optional response object.
        """

    async def process(self, req: Request, res: Response) -> Response:
        """Process an incoming request.

        - Call `before_dispatch()`. If a response is returned here, no
        further processing is performed.
        - Call the underlying HTTP `app`.
        - Call `after_dispatch()`.
        - Return the response.

        Note: this is aliased to `__call__()`, which means middleware
        instances are callable.

        # Parameters
        req (Request): a Request object.
        res (Response): a Response object.

        # Returns
        res (Response): a Response object.
        """
        before_res: Optional[Response] = await call_async(  # type: ignore
            self.before_dispatch, req, res
        )
        if before_res:
            return before_res

        res = await self.app(req, res)

        res = (
            await call_async(self.after_dispatch, req, res)  # type: ignore
            or res
        )

        return res

    __call__ = process



"""Package setup."""

import setuptools

description = "A modern Python web framework filled with asynchronous salsa."

with open("README.md", "r") as readme:
    long_description = readme.read()

GITHUB = "https://github.com/bocadilloproject/bocadillo"
DOCS = "https://bocadilloproject.github.io"
CHANGELOG = f"{GITHUB}/blob/master/CHANGELOG.md"

setuptools.setup(
    name="bocadillo",
    version="0.12.1",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["bocadillo"],
    package_data={"bocadillo": ["assets/*"]},
    install_requires=[
        "starlette>=0.11",
        "uvicorn>=0.3.26",
        "jinja2>=2.10",
        "whitenoise",
        "requests",
        "parse",
        "python-multipart",
        "websockets>=6.0",
    ],
    extras_require={"files": ["aiofiles"]},
    python_requires=">=3.6",
    url=DOCS,
    project_urls={
        "Source": GITHUB,
        "Documentation": DOCS,
        "Changelog": CHANGELOG,
    },
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



from typing import Sequence

from .applications import App


class RecipeBase:
    """Definition of the recipe interface."""

    def __init__(self, prefix: str):
        assert prefix.startswith("/"), "recipe prefix must start with '/'"
        self.prefix = prefix

    def apply(self, app: App, root: str = ""):
        """Apply the recipe to an application.

        Should be implemented by subclasses.

        # Parameters
        app (App): an application instance.
        root (str): a root URL path.
        """
        raise NotImplementedError


class Recipe(App):
    """A grouping of capabilities that can be merged back into an application.

    # Parameters

    name (str):
        A name for the recipe.
    prefix (str):
        The path prefix where the recipe will be mounted.
        Defaults to `"/" + name`.
    """

    def __init__(self, name: str, prefix: str = None, **kwargs):
        super().__init__(name=name, **kwargs)

        if prefix is None:
            prefix = f"/{name}"
        assert prefix.startswith("/"), "recipe prefix must start with '/'"

        self.prefix = prefix
        # DEPRECATED: 0.13.0
        self._templates_dir_given = "templates_dir" in kwargs

    def _get_own_url_for(self, name: str, **kwargs) -> str:
        return self.prefix + super()._get_own_url_for(name, **kwargs)

    def apply(self, app: App, root: str = ""):
        """Apply the recipe to an application."""
        app.mount(prefix=root + self.prefix, app=self)

        # DEPRECATED: 0.13.0
        # Look for templates where the app does, if not specified
        if not self._templates_dir_given:
            self.templates_dir = app.templates_dir  # type: ignore

    @classmethod
    def book(cls, *recipes: "Recipe", prefix: str) -> "RecipeBook":
        """Build a book of recipes.

        Shortcut for `RecipeBook(recipes, prefix)`.
        """
        return RecipeBook(recipes, prefix)


class RecipeBook(RecipeBase):
    """A composition of multiple recipes.

    # Parameters
    recipes (list): a list of `Recipe` objects.
    prefix (str):
        A prefix that will be prepended to all of the recipes' prefixes.
    """

    def __init__(self, recipes: Sequence[Recipe], prefix: str):
        super().__init__(prefix)
        self.recipes = recipes

    def apply(self, app: App, root: str = ""):
        """Apply the recipe book to an application."""
        for recipe in self.recipes:
            recipe.apply(app, root=root + self.prefix)



import os
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.middleware.wsgi import WSGIResponder
from starlette.routing import Lifespan
from starlette.testclient import TestClient
from uvicorn.config import get_logger
from uvicorn.main import run
from uvicorn.reloaders.statreload import StatReload

from .app_types import (
    _E,
    ASGIApp,
    ASGIAppInstance,
    ErrorHandler,
    EventHandler,
    Receive,
    Scope,
    Send,
)
from .compat import WSGIApp
from .constants import CONTENT_TYPE, DEFAULT_CORS_CONFIG
from .deprecation import deprecated
from .error_handlers import error_to_text
from .errors import HTTPError, HTTPErrorMiddleware, ServerErrorMiddleware
from .media import UnsupportedMediaType, get_default_handlers
from .meta import DocsMeta
from .request import Request
from .response import Response
from .routing import RoutingMixin
from .staticfiles import static
from .templates import TemplatesMixin

if TYPE_CHECKING:  # pragma: no cover
    from .recipes import Recipe


class App(TemplatesMixin, RoutingMixin, metaclass=DocsMeta):
    """The all-mighty application class.

    This class implements the [ASGI](https://asgi.readthedocs.io) protocol.

    # Example

    ```python
    >>> from bocadillo import App
    >>> app = App()
    ```

    # Parameters

    name (str):
        An optional name for the app.
    static_dir (str):
        The name of the directory containing static files, relative to
        the application entry point. Set to `None` to not serve any static
        files.
        Defaults to `"static"`.
    static_root (str):
        The path prefix for static assets.
        Defaults to `"static"`.
    allowed_hosts (list of str, optional):
        A list of hosts which the server is allowed to run at.
        If the list contains `"*"`, any host is allowed.
        Defaults to `["*"]`.
    enable_cors (bool):
        If `True`, Cross Origin Resource Sharing will be configured according
        to `cors_config`. Defaults to `False`.
        See also [CORS](../guides/http/middleware.md#cors).
    cors_config (dict):
        A dictionary of CORS configuration parameters.
        Defaults to `dict(allow_origins=[], allow_methods=["GET"])`.
    enable_hsts (bool):
        If `True`, enable HSTS (HTTP Strict Transport Security) and automatically
        redirect HTTP traffic to HTTPS.
        Defaults to `False`.
        See also [HSTS](../guides/http/middleware.md#hsts).
    enable_gzip (bool):
        If `True`, enable GZip compression and automatically
        compress responses for clients that support it.
        Defaults to `False`.
        See also [GZip](../guides/http/middleware.md#gzip).
    gzip_min_size (int):
        If specified, compress only responses that
        have more bytes than the specified value.
        Defaults to `1024`.
    media_type (str):
        Determines how values given to `res.media` are serialized.
        Can be one of the supported media types.
        Defaults to `"application/json"`.
        See also [Media](../guides/http/media.md).

    # Attributes
    media_handlers (dict):
        The dictionary of media handlers.
        You can access, edit or replace this at will.
    """

    def __init__(
        self,
        name: str = None,
        *,
        static_dir: Optional[str] = "static",
        static_root: Optional[str] = "static",
        allowed_hosts: List[str] = None,
        enable_cors: bool = False,
        cors_config: dict = None,
        enable_hsts: bool = False,
        enable_gzip: bool = False,
        gzip_min_size: int = 1024,
        media_type: str = CONTENT_TYPE.JSON,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.name = name

        # Debug mode defaults to `False` but it can be set in `.run()`.
        self._debug = False

        # Base ASGI app
        self.asgi = self.dispatch

        # Mounted (children) apps
        self._prefix_to_app: Dict[str, Any] = {}
        self._name_to_prefix_and_app: Dict[str, Tuple[str, App]] = {}

        # Test client
        self.client = self.build_client()

        # Static files
        if static_dir is not None:
            if static_root is None:
                static_root = static_dir
            self.mount(static_root, static(static_dir))

        # Media
        self.media_handlers = get_default_handlers()
        self._media_type = ""
        self.media_type = media_type

        # HTTP middleware
        self.exception_middleware = HTTPErrorMiddleware(
            self.http_router, debug=self._debug
        )
        self.server_error_middleware = ServerErrorMiddleware(
            self.exception_middleware, handler=error_to_text, debug=self._debug
        )
        self.add_error_handler(HTTPError, error_to_text)

        # Lifespan middleware
        self._lifespan = Lifespan()

        # ASGI middleware
        if allowed_hosts is None:
            allowed_hosts = ["*"]
        self.add_asgi_middleware(
            TrustedHostMiddleware, allowed_hosts=allowed_hosts
        )
        if enable_cors:
            cors_config = {**DEFAULT_CORS_CONFIG, **(cors_config or {})}
            self.add_asgi_middleware(CORSMiddleware, **cors_config)
        if enable_hsts:
            self.add_asgi_middleware(HTTPSRedirectMiddleware)
        if enable_gzip:
            self.add_asgi_middleware(GZipMiddleware, minimum_size=gzip_min_size)

    @property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, debug: bool):
        self._debug = debug
        self.exception_middleware.debug = debug
        self.server_error_middleware.debug = debug

    @property
    def media_type(self) -> str:
        """The media type configured when instanciating the application."""
        return self._media_type

    @media_type.setter
    def media_type(self, media_type: str):
        if media_type not in self.media_handlers:
            raise UnsupportedMediaType(media_type, handlers=self.media_handlers)
        self._media_type = media_type

    def build_client(self, **kwargs) -> TestClient:
        return TestClient(self, **kwargs)

    def get_template_globals(self):
        # DEPRECATED: 0.13.0
        return {"url_for": self.url_for}

    def url_for(self, name: str, **kwargs) -> str:
        # Implement route name lookup accross sub-apps.
        try:
            return super().url_for(name, **kwargs)
        except HTTPError as exc:
            app_name, _, name = name.partition(":")

            if not name:
                # No app name given.
                raise exc from None

            return self._url_for_app(app_name, name, **kwargs)

    def _url_for_app(self, app_name: str, name: str, **kwargs) -> str:
        if app_name == self.name:
            # NOTE: this allows to reference this app's routes in
            # both with or without the namespace.
            return self._get_own_url_for(name, **kwargs)

        try:
            prefix, app = self._name_to_prefix_and_app[app_name]
        except KeyError as key_exc:
            raise HTTPError(404) from key_exc
        else:
            return prefix + app.url_for(name, **kwargs)

    def _get_own_url_for(self, name: str, **kwargs) -> str:
        # NOTE: recipes hook into this method to prepend their
        # prefix to the URL.
        return super().url_for(name, **kwargs)

    def mount(self, prefix: str, app: Union["App", ASGIApp, WSGIApp]):
        """Mount another WSGI or ASGI app at the given prefix.

        # Parameters
        prefix (str): A path prefix where the app should be mounted, e.g. `"/myapp"`.
        app: An object implementing [WSGI](https://wsgi.readthedocs.io) or [ASGI](https://asgi.readthedocs.io) protocol.
        """
        if not prefix.startswith("/"):
            prefix = "/" + prefix

        self._prefix_to_app[prefix] = app

        if isinstance(app, App) and app.name is not None:
            self._name_to_prefix_and_app[app.name] = (prefix, app)

    def recipe(self, recipe: "Recipe"):
        """Apply a recipe.

        # Parameters
        recipe (Recipe or RecipeBook):
            a recipe to be applied to the application.

        # See Also
        - [Recipes](../guides/agnostic/recipes.md)
        """
        recipe.apply(self)

    def add_error_handler(self, exception_cls: Type[_E], handler: ErrorHandler):
        """Register a new error handler.

        # Parameters
        exception_cls (Exception class):
            The type of exception that should be handled.
        handler (callable):
            The actual error handler, which is called when an instance of
            `exception_cls` is caught.
            Should accept a request, response and exception parameters.
        """
        self.exception_middleware.add_exception_handler(exception_cls, handler)

    def error_handler(self, exception_cls: Type[Exception]):
        """Register a new error handler (decorator syntax).

        # See Also
        - [add_error_handler](#add-error-handler)
        """

        def wrapper(handler):
            self.add_error_handler(exception_cls, handler)
            return handler

        return wrapper

    def add_middleware(self, middleware_cls, **kwargs):
        """Register a middleware class.

        # Parameters

        middleware_cls (Middleware class):
            A subclass of `bocadillo.Middleware`.

        # See Also
        - [Middleware](../guides/http/middleware.md)
        """
        self.exception_middleware.app = middleware_cls(
            self.exception_middleware.app, **kwargs
        )

    def add_asgi_middleware(self, middleware_cls, *args, **kwargs):
        """Register an ASGI middleware class.

        # Parameters
        middleware_cls (Middleware class):
            A class that complies with the ASGI specification.

        # See Also
        - [ASGI middleware](../guides/agnostic/asgi-middleware.md)
        - [ASGI](https://asgi.readthedocs.io)
        """
        self.asgi = middleware_cls(self.asgi, *args, **kwargs)

    def on(self, event: str, handler: Optional[EventHandler] = None):
        """Register an event handler.

        # Parameters
        event (str):
            Either `"startup"` (when the server boots) or `"shutdown"`
            (when the server stops).
        handler (callback, optional):
            The event handler. If not given, this should be used as a
            decorator.

        # Example

        ```python
        @app.on("startup")
        async def startup():
            pass

        async def shutdown():
            pass

        app.on("shutdown", shutdown)
        ```
        """
        if handler is None:

            def register(func):
                self._lifespan.add_event_handler(event, func)
                return func

            return register
        else:
            self._lifespan.add_event_handler(event, handler)
            return handler

    async def dispatch_http(self, receive: Receive, send: Send, scope: Scope):
        req = Request(scope, receive)
        res = Response(
            req,
            media_type=self.media_type,
            media_handler=self.media_handlers[self.media_type],
        )
        response = await self.server_error_middleware(req, res)

        await response(receive, send)

        # Re-raise the exception to allow the server to log the error
        # and for the test client to optionally re-raise it too.
        self.server_error_middleware.raise_if_exception()

    async def dispatch_websocket(
        self, receive: Receive, send: Send, scope: Scope
    ):
        await self.websocket_router(scope, receive, send)

    def dispatch(self, scope: Scope) -> ASGIAppInstance:
        if scope["type"] == "websocket":
            return partial(self.dispatch_websocket, scope=scope)
        assert scope["type"] == "http"
        return partial(self.dispatch_http, scope=scope)

    def __call__(self, scope: Scope) -> ASGIAppInstance:
        if scope["type"] == "lifespan":
            return self._lifespan(scope)

        path: str = scope["path"]

        # Return a sub-mounted extra app, if found
        for prefix, app in self._prefix_to_app.items():
            if not path.startswith(prefix):
                continue
            # Remove prefix from path so that the request is made according
            # to the mounted app's point of view.
            scope["path"] = path[len(prefix) :]
            try:
                return app(scope)
            except TypeError:
                return WSGIResponder(app, scope)

        return self.asgi(scope)

    def run(
        self,
        host: str = None,
        port: int = None,
        debug: bool = False,
        log_level: str = "info",
        _run: Callable = None,
        **kwargs,
    ):
        """Serve the application using [uvicorn](https://www.uvicorn.org).

        # Parameters

        host (str):
            The host to bind to.
            Defaults to `"127.0.0.1"` (localhost).
            If not given and `$PORT` is set, `"0.0.0.0"` will be used to
            serve to all known hosts.
        port (int):
            The port to bind to.
            Defaults to `8000` or (if set) the value of the `$PORT` environment
            variable.
        debug (bool):
            Whether to serve the application in debug mode. Defaults to `False`.
        log_level (str):
            A logging level for the debug logger. Must be a logging level
            from the `logging` module. Defaults to `"info"`.
        kwargs (dict):
            Extra keyword arguments that will be passed to the Uvicorn runner.

        # See Also
        - [Configuring host and port](../guides/app.md#configuring-host-and-port)
        - [Debug mode](../guides/app.md#debug-mode)
        - [Uvicorn settings](https://www.uvicorn.org/settings/) for all
        available keyword arguments.
        """
        if _run is None:  # pragma: no cover
            _run = run

        if "PORT" in os.environ:
            port = int(os.environ["PORT"])
            if host is None:
                host = "0.0.0.0"

        if host is None:
            host = "127.0.0.1"

        if port is None:
            port = 8000

        if debug:
            self.debug = True
            reloader = StatReload(get_logger(log_level))
            kwargs = {
                "app": self,
                "host": host,
                "port": port,
                "log_level": log_level,
                "debug": self.debug,
                **kwargs,
            }
            reloader.run(run, kwargs)
        else:
            _run(self, host=host, port=port, **kwargs)


@deprecated(
    since="0.12",
    removal="0.13",
    alternative=("bocadillo.App", "/api/applications.md#App"),
    warn_on_instanciate=True,
)
class API(App):
    """The all-mighty API class. An alias to `App`, nothing more."""



from os.path import exists

from whitenoise import WhiteNoise

from .compat import WSGIApp, empty_wsgi_app


def static(directory: str) -> WSGIApp:
    """Return a WSGI app that serves static files under the given directory.

    Powered by WhiteNoise.

    # Parameters
    directory (str):
        the path to a directory from where static files should be served.
        If the directory does not exist, no files will be served.

    # Returns
    app (WSGIApp): a WSGI-compliant application.

    # See Also
    - [WhiteNoise](http://whitenoise.evans.io)
    - [WSGI](https://wsgi.readthedocs.io)
    """
    app = WhiteNoise(empty_wsgi_app())
    if exists(directory):
        app.add_files(directory)
    return app
